'use strict';

const vfs = require('@previewjs/vfs');
const react = require('@vitejs/plugin-react');
const path = require('path');
const ts = require('typescript');
const csf3 = require('@previewjs/csf3');
const serializableValues = require('@previewjs/serializable-values');
const typeAnalyzer = require('@previewjs/type-analyzer');

function detectPropTypes(sourceFile, name) {
  for (const statement of sourceFile.statements) {
    if (ts.isExpressionStatement(statement) && ts.isBinaryExpression(statement.expression) && statement.expression.operatorToken.kind === ts.SyntaxKind.EqualsToken && ts.isPropertyAccessExpression(statement.expression.left) && ts.isIdentifier(statement.expression.left.expression)) {
      const componentName = statement.expression.left.expression.text;
      if (componentName !== name) {
        continue;
      }
      if (statement.expression.left.name.text === "propTypes") {
        return statement.expression.right;
      }
    } else if (ts.isClassDeclaration(statement) && statement.name) {
      const componentName = statement.name.text;
      if (componentName !== name) {
        continue;
      }
      for (const classMember of statement.members) {
        if (ts.isPropertyDeclaration(classMember) && classMember.initializer && classMember.name.getText() === "propTypes") {
          return classMember.initializer;
        }
      }
    }
  }
  return null;
}

function analyzeReactComponent(typeResolver, absoluteFilePath, componentName, signature) {
  const sourceFile = typeResolver.sourceFile(absoluteFilePath);
  let propTypes = null;
  if (sourceFile) {
    propTypes = detectPropTypes(sourceFile, componentName);
  }
  const resolved = computePropsType(typeResolver, signature, propTypes);
  return {
    propsType: resolved.type,
    types: { ...resolved.collected }
  };
}
function computePropsType(typeResolver, signature, propTypes) {
  if (propTypes) {
    return computePropsTypeFromPropTypes(typeResolver, propTypes);
  }
  return computePropsTypeFromSignature(typeResolver, signature);
}
function computePropsTypeFromSignature(typeResolver, signature) {
  const firstParam = signature.getParameters()[0];
  if (!firstParam) {
    return {
      type: typeAnalyzer.EMPTY_OBJECT_TYPE,
      collected: {}
    };
  }
  if (!firstParam.valueDeclaration) {
    return {
      type: typeAnalyzer.UNKNOWN_TYPE,
      collected: {}
    };
  }
  const type = typeResolver.checker.getTypeOfSymbolAtLocation(
    firstParam,
    firstParam.valueDeclaration
  );
  try {
    let { type: propsType, collected } = typeResolver.resolveType(type);
    [propsType] = typeAnalyzer.dereferenceType(propsType, collected, []);
    stripUnusedProps:
      if (propsType.kind === "object" && ts.isParameter(firstParam.valueDeclaration)) {
        if (ts.isObjectBindingPattern(firstParam.valueDeclaration.name)) {
          const bindingPattern = firstParam.valueDeclaration.name;
          const usedProps = /* @__PURE__ */ new Set();
          const propsWithDefault = /* @__PURE__ */ new Set();
          for (const element of bindingPattern.elements) {
            if (element.dotDotDotToken) {
              break stripUnusedProps;
            }
            const elementName = element.propertyName || element.name;
            if (!ts.isIdentifier(elementName)) {
              break stripUnusedProps;
            }
            usedProps.add(elementName.text);
            if (element.initializer) {
              propsWithDefault.add(elementName.text);
            }
          }
          propsType = typeAnalyzer.objectType(
            Object.fromEntries(
              Object.entries(propsType.fields).filter(([key]) => usedProps.has(key)).map(([key, type2]) => [
                key,
                typeAnalyzer.maybeOptionalType(type2, propsWithDefault.has(key))
              ])
            )
          );
        }
      }
    return { type: propsType, collected };
  } catch (e) {
    console.warn(
      `Unable to resolve props type for ${typeResolver.checker.typeToString(
        type
      )}`,
      e
    );
  }
  return {
    type: typeAnalyzer.UNKNOWN_TYPE,
    collected: {}
  };
}
function computePropsTypeFromPropTypes(typeResolver, propTypes) {
  const type = typeResolver.checker.getTypeAtLocation(propTypes);
  const fields = {};
  let collected = {};
  for (const property of type.getProperties()) {
    fields[property.name] = (() => {
      if (!property.valueDeclaration) {
        return typeAnalyzer.UNKNOWN_TYPE;
      }
      const propertyType = typeResolver.checker.getTypeAtLocation(
        property.valueDeclaration
      );
      const typeArguments = typeResolver.resolveTypeArguments(propertyType);
      const fieldType = typeArguments.types[0];
      if (fieldType) {
        collected = {
          ...collected,
          ...typeArguments.collected
        };
        return typeAnalyzer.maybeOptionalType(
          fieldType,
          propertyType.symbol.name === "Requireable"
        );
      }
      return typeAnalyzer.UNKNOWN_TYPE;
    })();
  }
  const resolved = typeAnalyzer.objectType(fields);
  return {
    type: resolved,
    collected: typeAnalyzer.stripUnusedTypes(collected, resolved)
  };
}

function extractReactComponents(resolver, absoluteFilePath) {
  const sourceFile = resolver.sourceFile(absoluteFilePath);
  if (!sourceFile) {
    return [];
  }
  const functions = [];
  for (const statement of sourceFile.statements) {
    if (ts.isExportAssignment(statement)) {
      if (ts.isIdentifier(statement.expression)) {
        continue;
      }
      functions.push(["default", statement, statement.expression]);
    } else if (ts.isVariableStatement(statement)) {
      for (const declaration of statement.declarationList.declarations) {
        if (!ts.isIdentifier(declaration.name) || !declaration.initializer) {
          continue;
        }
        functions.push([
          declaration.name.text,
          statement,
          declaration.initializer
        ]);
      }
    } else if (ts.isFunctionDeclaration(statement)) {
      const isDefaultExport = !!statement.modifiers?.find(
        (m) => m.kind === ts.SyntaxKind.ExportKeyword
      ) && !!statement.modifiers?.find(
        (m) => m.kind === ts.SyntaxKind.DefaultKeyword
      );
      const name = statement.name?.text;
      if (isDefaultExport || name) {
        functions.push([name || "default", statement, statement]);
      }
    } else if (ts.isClassDeclaration(statement) && statement.name) {
      functions.push([statement.name.text, statement, statement]);
    }
  }
  const storiesDefaultComponent = csf3.extractDefaultComponent(sourceFile);
  const resolvedStoriesComponent = storiesDefaultComponent ? csf3.resolveComponent(resolver.checker, storiesDefaultComponent) : null;
  const components = [];
  const args = typeAnalyzer.helpers.extractArgs(sourceFile);
  const nameToExportedName = typeAnalyzer.helpers.extractExportedNames(sourceFile);
  function extractComponentTypeInfo(node, name) {
    const storyArgs = args[name];
    const isExported = name === "default" || !!nameToExportedName[name];
    if (storiesDefaultComponent && storyArgs && isExported) {
      return {
        kind: "story",
        args: {
          start: storyArgs.getStart(),
          end: storyArgs.getEnd(),
          value: serializableValues.parseSerializableValue(storyArgs)
        },
        associatedComponent: resolvedStoriesComponent
      };
    }
    const signature = extractComponentSignature(resolver.checker, node);
    if (signature) {
      return {
        kind: "component",
        exported: isExported,
        analyze: async () => analyzeReactComponent(resolver, absoluteFilePath, name, signature)
      };
    }
    return null;
  }
  for (const [name, statement, node] of functions) {
    const info = extractComponentTypeInfo(node, name);
    if (info) {
      components.push({
        absoluteFilePath,
        name,
        offsets: [[statement.getStart(), statement.getEnd()]],
        info
      });
    }
  }
  return [...components, ...csf3.extractCsf3Stories(resolver, sourceFile)];
}
function extractComponentSignature(checker, node) {
  const type = checker.getTypeAtLocation(node);
  for (const callSignature of type.getCallSignatures()) {
    if (isValidComponentReturnType(callSignature.getReturnType())) {
      return callSignature;
    }
  }
  if (type.symbol) {
    const classType = checker.getTypeOfSymbolAtLocation(type.symbol, node);
    for (const constructSignature of classType.getConstructSignatures()) {
      const returnType = constructSignature.getReturnType();
      if (returnType.getProperty("render")) {
        return constructSignature;
      }
    }
  }
  return null;
}
function isValidComponentReturnType(type) {
  if (isJsxElement(type)) {
    return true;
  }
  return false;
}
const jsxElementTypes = /* @__PURE__ */ new Set(["Element", "ReactElement"]);
function isJsxElement(type) {
  if (type.isUnion()) {
    for (const subtype of type.types) {
      if (isJsxElement(subtype)) {
        return true;
      }
    }
  }
  return jsxElementTypes.has(type.symbol?.getEscapedName().toString());
}

const reactImportRegExp = /import (\* as )?React[ ,]/;
function reactImportsPlugin() {
  return {
    name: "previewjs:react-js-imports",
    async transform(code, id) {
      if (id.includes("node_modules") || path.extname(id) !== ".js" || reactImportRegExp.test(code)) {
        return code;
      }
      return `import React from 'react';${code}`;
    }
  };
}

const REACT_SPECIAL_TYPES = {
  Component: typeAnalyzer.NODE_TYPE,
  ComponentClass: typeAnalyzer.functionType(typeAnalyzer.NODE_TYPE),
  FunctionComponent: typeAnalyzer.functionType(typeAnalyzer.NODE_TYPE),
  ReactElement: typeAnalyzer.NODE_TYPE,
  ReactNode: typeAnalyzer.NODE_TYPE,
  ReactNodeArray: typeAnalyzer.arrayType(typeAnalyzer.NODE_TYPE),
  ReactPortal: typeAnalyzer.NODE_TYPE
};

const reactFrameworkPlugin = {
  isCompatible: async (dependencies) => {
    const version = await dependencies["react"]?.readInstalledVersion();
    if (!version) {
      return false;
    }
    return parseInt(version) >= 16;
  },
  async create({ rootDirPath, dependencies }) {
    const previewDirPath = path.join(__dirname, "..", "preview");
    return {
      pluginApiVersion: 3,
      name: "@previewjs/plugin-react",
      defaultWrapperPath: "__previewjs__/Wrapper.tsx",
      previewDirPath,
      specialTypes: REACT_SPECIAL_TYPES,
      tsCompilerOptions: {
        jsx: ts.JsxEmit.ReactJSX,
        jsxImportSource: "react"
      },
      transformReader: (reader) => vfs.createStackedReader([
        reader,
        vfs.createFileSystemReader({
          mapping: {
            from: path.join(previewDirPath, "types"),
            to: path.join(rootDirPath, "node_modules", "@types")
          },
          watch: false
        })
      ]),
      detectComponents: async (typeAnalyzer, absoluteFilePaths) => {
        const resolver = typeAnalyzer.analyze(absoluteFilePaths);
        const components = [];
        for (const absoluteFilePath of absoluteFilePaths) {
          components.push(
            ...extractReactComponents(resolver, absoluteFilePath)
          );
        }
        return components;
      },
      viteConfig: () => {
        return {
          resolve: {
            alias: {
              "react-native": "react-native-web"
            }
          },
          plugins: [
            reactImportsPlugin(),
            react(),
            {
              name: "previewjs:update-react-import",
              async transform(code, id) {
                if (!id.endsWith("__previewjs_internal__/renderer/index.tsx")) {
                  return;
                }
                const reactVersion = parseInt(
                  await dependencies["react"]?.readInstalledVersion() || "0"
                );
                return code.replace(
                  /__PREVIEWJS_PLUGIN_REACT_IMPORT_PATH__/g,
                  reactVersion >= 18 ? "./render-18" : "./render-16"
                );
              }
            },
            {
              name: "previewjs:disable-react-hmr",
              async transform(code, id) {
                if (!id.endsWith(".jsx") && !id.endsWith(".tsx")) {
                  return null;
                }
                return code.replace(/import\.meta/g, "({})");
              }
            }
          ],
          define: {
            "process.env.RUNNING_INSIDE_PREVIEWJS": "1"
          }
        };
      }
    };
  }
};

module.exports = reactFrameworkPlugin;
